using System;
using System.Diagnostics;
using System.Threading.Tasks;
using System.Windows.Threading;

namespace FPLab.Utils
{
    public delegate Task AsyncHandler(object sender, EventArgs args);

    public class AutoLoadDispatcherTimer
    {
        private readonly DispatcherTimer _internalTimer;
        private int _availableProcessors;
        private Stopwatch _intervalStopwatch;

        private int _operationsCount;
        private long _operationsDuration;

        /// <summary>
        /// Initializes a new instance of the AutoLoadDispatcherTimer class.
        /// </summary>
        public AutoLoadDispatcherTimer()
        {
            _internalTimer = new DispatcherTimer();
            _internalTimer.Tick += AutoLoadCallback;
            Initialize();
        }

        /// <summary>
        /// Initializes a new instance of the AutoLoadDispatcherTimer class.
        /// </summary>
        /// <param name="priority">The priority at which to invoke the timer.</param>
        public AutoLoadDispatcherTimer(DispatcherPriority priority)
        {
            _internalTimer = new DispatcherTimer(priority);
            _internalTimer.Tick += AutoLoadCallback;
            Initialize();
        }

        /// <summary>
        /// Initializes a new instance of the AutoLoadDispatcherTimer class.
        /// </summary>
        /// <param name="priority">The priority at which to invoke the timer.</param>
        /// <param name="dispatcher">The dispatcher the timer is associated with.</param>
        public AutoLoadDispatcherTimer(DispatcherPriority priority, Dispatcher dispatcher)
        {
            _internalTimer = new DispatcherTimer(priority, dispatcher);
            _internalTimer.Tick += AutoLoadCallback;
            Initialize();
        }

        /// <summary>
        /// Initializes a new instance of the AutoLoadDispatcherTimer class.
        /// </summary>
        /// <param name="interval">The period of time between ticks.</param>
        /// <param name="priority">The priority at which to invoke the timer.</param>
        /// <param name="callback">The event handler to call when the AutoLoadDispatcherTimer.Tick event occurs.</param>
        /// <param name="dispatcher">The dispatcher the timer is associated with</param>
        public AutoLoadDispatcherTimer(TimeSpan interval, DispatcherPriority priority, AsyncHandler callback,
                                       Dispatcher dispatcher)
        {
            Tick += callback;
            _internalTimer = new DispatcherTimer(interval, priority, AutoLoadCallback, dispatcher);
            Initialize();
        }

        /// <summary>
        /// Gets the System.Windows.Threading.Dispatcher associated with this AutoLoadDispatcherTimer.
        /// </summary>
        /// <value>The dispatcher.</value>
        public Dispatcher Dispatcher
        {
            get { return _internalTimer.Dispatcher; }
        }

        /// <summary>
        /// Gets or sets a value that indicates whether the timer is running.
        /// </summary>
        /// <value>Returns true if the timer is enabled; otherwise, false. The default is false.</value>
        public bool IsEnabled
        {
            get { return _internalTimer.IsEnabled; }
            set { _internalTimer.IsEnabled = value; }
        }

        /// <summary>
        /// Gets or sets the period of time between timer ticks.
        /// </summary>
        /// <value>The period of time between ticks. The default is 00:00:00.</value>
        public TimeSpan Interval
        {
            get { return _internalTimer.Interval; }
            set { _internalTimer.Interval = value; }
        }

        /// <summary>
        /// Gets or sets a user-defined data object.
        /// </summary>
        /// <value>The user-defined data. The default is null.</value>
        public object Tag
        {
            get { return _internalTimer.Tag; }
            set { _internalTimer.Tag = value; }
        }

        /// <summary>
        /// Gets the number of operations performed per second.
        /// </summary>
        /// <value>Number of operations performed per second.</value>
        public double OperationsPerSecond { get; private set; }

        private void Initialize()
        {
            IntervalMargin = 10;
            _intervalStopwatch = Stopwatch.StartNew();
            _availableProcessors = GetAvailableProcessors();
        }

        /// <summary>
        /// Starts the AutoLoadDispatcherTimer.
        /// </summary>
        public void Start()
        {
            _internalTimer.Start();
        }

        /// <summary>
        /// Stops the AutoLoadDispatcherTimer.
        /// </summary>
        public void Stop()
        {
            _internalTimer.Stop();
        }

        /// <summary>
        /// Occurs when the timer interval has elapsed.
        /// </summary>
        public event AsyncHandler Tick;

        /// <summary>
        /// Invokes asynchronously the Tick event.
        /// </summary>
        /// <returns>Awaitable object for async operation.</returns>
        protected virtual async Task OnTick()
        {
            AsyncHandler handler = Tick;
            if (handler != null) await handler(this, EventArgs.Empty);
        }

        private async void AutoLoadCallback(object sender, EventArgs eventArgs)
        {
            _intervalStopwatch.Restart();

            await OnTick();
            _operationsCount++;

            _intervalStopwatch.Stop();

            // calculate the interval, so the scanning will use all available cores

            long interval = (_intervalStopwatch.ElapsedMilliseconds/_availableProcessors) + IntervalMargin;
            _internalTimer.Interval = TimeSpan.FromMilliseconds(interval);

            _operationsDuration += _intervalStopwatch.ElapsedMilliseconds;

            if (_operationsDuration >= 1000)
            {
                OperationsPerSecond = (_operationsCount/(double) _operationsDuration)*1000.0;

                _operationsCount = 0;
                _operationsDuration = 0;
            }
        }

        /// <summary>
        /// Gets or sets Time Margin in miliseconds that will always be added to the interval.
        /// This is used to ensure that Tick will not be queued.
        /// </summary>
        /// <value>Time Margin in miliseconds. Default value is 10.</value>
        public long IntervalMargin { get; set; }

        private static int GetAvailableProcessors()
        {
            int availableProcessors = Environment.ProcessorCount - 1;
            if (availableProcessors <= 0)
                availableProcessors = 1;

            return availableProcessors;
        }
    }
}